@@ -24,7 +24,7 @@
 
 -import(couch_httpd,
     [send_json/2,send_json/3,send_json/4,send_method_not_allowed/2,
-    start_json_response/2,send_chunk/2,end_json_response/1,
+    start_json_response/2,send_chunk/2,last_chunk/1,end_json_response/1,
     start_chunked_response/3, send_error/4]).
 
 % httpd global handlers
@@ -38,7 +38,15 @@ handle_welcome_req(Req, _) ->
     send_method_not_allowed(Req, "GET,HEAD").
 
 handle_favicon_req(#httpd{method='GET'}=Req, DocumentRoot) ->
-    couch_httpd:serve_file(Req, "favicon.ico", DocumentRoot);
+    {{Year,Month,Day},Time} = erlang:localtime(),
+    OneYearFromNow = {{Year+1,Month,Day},Time},
+    CachingHeaders = [
+        %favicon should expire a year from now
+        {"Cache-Control", "public, max-age=31536000"},
+        {"Expires", httpd_util:rfc1123_date(OneYearFromNow)}
+    ],
+    couch_httpd:serve_file(Req, "favicon.ico", DocumentRoot, CachingHeaders);
+
 handle_favicon_req(Req, _) ->
     send_method_not_allowed(Req, "GET,HEAD").
 
@@ -82,10 +90,14 @@ handle_replicate_req(#httpd{method='POST'}=Req) ->
     try couch_rep:replicate(PostBody, Req#httpd.user_ctx) of
     {ok, {continuous, RepId}} ->
         send_json(Req, 202, {[{ok, true}, {<<"_local_id">>, RepId}]});
+    {ok, {cancelled, RepId}} ->
+        send_json(Req, 200, {[{ok, true}, {<<"_local_id">>, RepId}]});
     {ok, {JsonResults}} ->
         send_json(Req, {[{ok, true} | JsonResults]});
     {error, {Type, Details}} ->
         send_json(Req, 500, {[{error, Type}, {reason, Details}]});
+    {error, not_found} ->
+        send_json(Req, 404, {[{error, not_found}]});
     {error, Reason} ->
         send_json(Req, 500, {[{error, Reason}]})
     catch
@@ -106,9 +118,7 @@ handle_restart_req(Req) ->
 
 handle_uuids_req(#httpd{method='GET'}=Req) ->
     Count = list_to_integer(couch_httpd:qs_value(Req, "count", "1")),
-    % generate the uuids
-    UUIDs = [ couch_util:new_uuid() || _ <- lists:seq(1,Count)],
-    % send a JSON response
+    UUIDs = [couch_uuids:new() || _ <- lists:seq(1, Count)],
     Etag = couch_httpd:make_etag(UUIDs),
     couch_httpd:etag_respond(Req, Etag, fun() ->
         CacheBustingHeaders = [
@@ -205,7 +215,7 @@ handle_log_req(#httpd{method='GET'}=Req) ->
         {"Content-Length", integer_to_list(length(Chunk))}
     ]),
     send_chunk(Resp, Chunk),
-    send_chunk(Resp, "");
+    last_chunk(Resp);
 handle_log_req(Req) ->
     send_method_not_allowed(Req, "GET").
 
