@@ -19,30 +19,33 @@
 oauth_authentication_handler(#httpd{mochi_req=MochiReq}=Req) ->
     serve_oauth(Req, fun(URL, Params, Consumer, Signature) ->
         AccessToken = proplists:get_value("oauth_token", Params),
-        TokenSecret = couch_config:get("oauth_token_secrets", AccessToken),
-        case oauth:verify(Signature, atom_to_list(MochiReq:get(method)), URL, Params, Consumer, TokenSecret) of
-            true ->
-                set_user_ctx(Req, AccessToken);
-            false ->
-                Req
+        case couch_config:get("oauth_token_secrets", AccessToken) of
+            undefined -> 
+                couch_httpd:send_error(Req, 400, <<"invalid_token">>,
+                    <<"Invalid OAuth token.">>);
+            TokenSecret ->
+                ?LOG_DEBUG("OAuth URL is: ~p", [URL]),
+                case oauth:verify(Signature, atom_to_list(MochiReq:get(method)), URL, Params, Consumer, TokenSecret) of
+                    true ->
+                        set_user_ctx(Req, AccessToken);
+                    false ->
+                        Req
+                end
         end
     end, true).
 
 % Look up the consumer key and get the roles to give the consumer
 set_user_ctx(Req, AccessToken) ->
-    DbName = couch_config:get("couch_httpd_auth", "authentication_db"),
-    couch_httpd_auth:ensure_users_db_exists(?l2b(DbName)),
-    case couch_db:open(?l2b(DbName), [{user_ctx, #user_ctx{roles=[<<"_admin">>]}}]) of
-        {ok, Db} ->
-            Name = ?l2b(couch_config:get("oauth_token_users", AccessToken)),
-            case couch_httpd_auth:get_user(Db, Name) of
-                nil -> Req;
-                User ->
-                    Roles = proplists:get_value(<<"roles">>, User, []),
-                    Req#httpd{user_ctx=#user_ctx{name=Name, roles=Roles}}
-            end;
-        _Else->
-            Req
+    % TODO move to db storage
+    Name = case couch_config:get("oauth_token_users", AccessToken) of
+        undefined -> throw({bad_request, unknown_oauth_token});
+        Value -> ?l2b(Value)
+    end,
+    case couch_httpd_auth:get_user(Name) of
+        nil -> Req;
+        User ->
+            Roles = proplists:get_value(<<"roles">>, User, []),
+            Req#httpd{user_ctx=#user_ctx{name=Name, roles=Roles}}
     end.
 
 % OAuth request_token
