@@ -13,20 +13,21 @@
 -module(couch_httpd_view).
 -include("couch_db.hrl").
 
--export([handle_view_req/2,handle_temp_view_req/2,handle_db_view_req/2]).
+-export([handle_view_req/3,handle_temp_view_req/2]).
 
 -export([get_stale_type/1, get_reduce_type/1, parse_view_params/3]).
--export([make_view_fold_fun/6, finish_view_fold/3, view_row_obj/3]).
+-export([make_view_fold_fun/6, finish_view_fold/4, view_row_obj/3]).
 -export([view_group_etag/2, view_group_etag/3, make_reduce_fold_funs/5]).
--export([design_doc_view/5, parse_bool_param/1]).
+-export([design_doc_view/5, parse_bool_param/1, doc_member/2]).
+-export([make_key_options/1, load_view/4]).
 
 -import(couch_httpd,
     [send_json/2,send_json/3,send_json/4,send_method_not_allowed/2,send_chunk/2,
     start_json_response/2, start_json_response/3, end_json_response/1,
     send_chunked_error/2]).
 
-design_doc_view(Req, Db, Id, ViewName, Keys) ->
-    DesignId = <<"_design/", Id/binary>>,
+design_doc_view(Req, Db, DName, ViewName, Keys) ->
+    DesignId = <<"_design/", DName/binary>>,
     Stale = get_stale_type(Req),
     Reduce = get_reduce_type(Req),
     Result = case couch_view:get_map_view(Db, DesignId, ViewName, Stale) of
@@ -53,11 +54,11 @@ design_doc_view(Req, Db, Id, ViewName, Keys) ->
     Result.
 
 handle_view_req(#httpd{method='GET',
-        path_parts=[_Db, _Design, DName, _View, ViewName]}=Req, Db) ->
+        path_parts=[_, _, DName, _, ViewName]}=Req, Db, _DDoc) ->
     design_doc_view(Req, Db, DName, ViewName, nil);
 
 handle_view_req(#httpd{method='POST',
-        path_parts=[_Db, _Design, DName, _View, ViewName]}=Req, Db) ->
+        path_parts=[_, _, DName, _, ViewName]}=Req, Db, _DDoc) ->
     {Fields} = couch_httpd:json_body_obj(Req),
     case proplists:get_value(<<"keys">>, Fields, nil) of
     nil ->
@@ -70,50 +71,7 @@ handle_view_req(#httpd{method='POST',
         throw({bad_request, "`keys` member must be a array."})
     end;
 
-handle_view_req(Req, _Db) ->
-    send_method_not_allowed(Req, "GET,POST,HEAD").
-
-handle_db_view_req(#httpd{method='GET',
-        path_parts=[_Db, _View, DName, ViewName]}=Req, Db) ->
-    QueryArgs = couch_httpd_view:parse_view_params(Req, nil, nil),
-    #view_query_args{
-        list = ListName
-    } = QueryArgs,
-    ?LOG_DEBUG("ici ~p", [ListName]),
-    case ListName of
-        nil -> couch_httpd_view:design_doc_view(Req, Db, DName, ViewName, nil);
-        _ ->
-            couch_httpd_show:handle_view_list(Req, DName, ListName, DName, ViewName, Db, nil)
-    end;
-
-handle_db_view_req(#httpd{method='POST',
-        path_parts=[_Db, _View, DName, ViewName]}=Req, Db) ->
-    QueryArgs = couch_httpd_view:parse_view_params(Req, nil, nil),
-    #view_query_args{
-        list = ListName
-    } = QueryArgs,
-    case ListName of
-    nil ->
-        {Fields} = couch_httpd:json_body_obj(Req),
-        case proplists:get_value(<<"keys">>, Fields, nil) of
-        nil ->
-            Fmt = "POST to view ~p/~p in database ~p with no keys member.",
-            ?LOG_DEBUG(Fmt, [DName, ViewName, Db]),
-            couch_httpd_view:design_doc_view(Req, Db, DName, ViewName, nil);
-        Keys when is_list(Keys) ->
-            couch_httpd_view:design_doc_view(Req, Db, DName, ViewName, Keys);
-        _ ->
-            throw({bad_request, "`keys` member must be a array."})
-        end;
-    _ ->
-        ReqBody = couch_httpd:body(Req),
-        {Props2} = ?JSON_DECODE(ReqBody),
-        Keys = proplists:get_value(<<"keys">>, Props2, nil),
-        couch_httpd_show:handle_view_list(Req#httpd{req_body=ReqBody},
-            DName, ListName, DName, ViewName, Db, Keys)
-    end;
-
-handle_db_view_req(Req, _Db) ->
+handle_view_req(Req, _Db, _DDoc) ->
     send_method_not_allowed(Req, "GET,POST,HEAD").
 
 handle_temp_view_req(#httpd{method='POST'}=Req, Db) ->
@@ -123,12 +81,18 @@ handle_temp_view_req(#httpd{method='POST'}=Req, Db) ->
     {DesignOptions} = proplists:get_value(<<"options">>, Props, {[]}),
     MapSrc = proplists:get_value(<<"map">>, Props),
     Keys = proplists:get_value(<<"keys">>, Props, nil),
+    Reduce = get_reduce_type(Req),
     case proplists:get_value(<<"reduce">>, Props, null) of
     null ->
         QueryArgs = parse_view_params(Req, Keys, map),
         {ok, View, Group} = couch_view:get_temp_map_view(Db, Language,
             DesignOptions, MapSrc),
         output_map_view(Req, View, Group, Db, QueryArgs, Keys);
+    _ when Reduce =:= false ->
+        QueryArgs = parse_view_params(Req, Keys, red_map),
+        {ok, View, Group} = couch_view:get_temp_map_view(Db, Language,
+            DesignOptions, MapSrc),
+        output_map_view(Req, View, Group, Db, QueryArgs, Keys);
     RedSrc ->
         QueryArgs = parse_view_params(Req, Keys, reduce),
         {ok, View, Group} = couch_view:get_temp_reduce_view(Db, Language,
@@ -142,65 +106,56 @@ handle_temp_view_req(Req, _Db) ->
 output_map_view(Req, View, Group, Db, QueryArgs, nil) ->
     #view_query_args{
         limit = Limit,
-        direction = Dir,
-        skip = SkipCount,
-        start_key = StartKey,
-        start_docid = StartDocId
+        skip = SkipCount
     } = QueryArgs,
     CurrentEtag = view_group_etag(Group, Db),
     couch_httpd:etag_respond(Req, CurrentEtag, fun() ->
         {ok, RowCount} = couch_view:get_row_count(View),
-        Start = {StartKey, StartDocId},
         FoldlFun = make_view_fold_fun(Req, QueryArgs, CurrentEtag, Db, RowCount, #view_fold_helper_funs{reduce_count=fun couch_view:reduce_to_count/1}),
-        FoldAccInit = {Limit, SkipCount, undefined, [], nil},
-        FoldResult = couch_view:fold(View, Start, Dir, FoldlFun, FoldAccInit),
-        finish_view_fold(Req, RowCount, FoldResult)
+        FoldAccInit = {Limit, SkipCount, undefined, []},
+        {ok, LastReduce, FoldResult} = couch_view:fold(View, 
+                FoldlFun, FoldAccInit, make_key_options(QueryArgs)),
+        finish_view_fold(Req, RowCount, 
+                couch_view:reduce_to_count(LastReduce), FoldResult)
     end);
 
 output_map_view(Req, View, Group, Db, QueryArgs, Keys) ->
     #view_query_args{
         limit = Limit,
-        direction = Dir,
-        skip = SkipCount,
-        start_docid = StartDocId
+        skip = SkipCount
     } = QueryArgs,
     CurrentEtag = view_group_etag(Group, Db, Keys),
     couch_httpd:etag_respond(Req, CurrentEtag, fun() ->
         {ok, RowCount} = couch_view:get_row_count(View),
-        FoldAccInit = {Limit, SkipCount, undefined, [], nil},
-        FoldResult = lists:foldl(
-            fun(Key, {ok, FoldAcc}) ->
-                Start = {Key, StartDocId},
+        FoldAccInit = {Limit, SkipCount, undefined, []},
+        {LastReduce, FoldResult} = lists:foldl(
+            fun(Key, {_, FoldAcc}) ->
                 FoldlFun = make_view_fold_fun(Req,
                     QueryArgs#view_query_args{
-                        start_key = Key,
-                        end_key = Key
                     }, CurrentEtag, Db, RowCount,
                     #view_fold_helper_funs{
                         reduce_count = fun couch_view:reduce_to_count/1
                     }),
-                couch_view:fold(View, Start, Dir, FoldlFun, FoldAcc)
-            end, {ok, FoldAccInit}, Keys),
-        finish_view_fold(Req, RowCount, FoldResult)
+                {ok, LastReduce, FoldResult} = couch_view:fold(View, FoldlFun, FoldAcc, 
+                    make_key_options(QueryArgs#view_query_args{start_key=Key, end_key=Key})),
+                {LastReduce, FoldResult}
+            end, {{[],[]}, FoldAccInit}, Keys),
+        finish_view_fold(Req, RowCount, couch_view:reduce_to_count(LastReduce), FoldResult)
     end).
 
 output_reduce_view(Req, Db, View, Group, QueryArgs, nil) ->
     #view_query_args{
-        start_key = StartKey,
-        end_key = EndKey,
         limit = Limit,
         skip = Skip,
-        direction = Dir,
-        start_docid = StartDocId,
-        end_docid = EndDocId,
         group_level = GroupLevel
     } = QueryArgs,
     CurrentEtag = view_group_etag(Group, Db),
     couch_httpd:etag_respond(Req, CurrentEtag, fun() ->
         {ok, GroupRowsFun, RespFun} = make_reduce_fold_funs(Req, GroupLevel, QueryArgs, CurrentEtag, #reduce_fold_helper_funs{}),
         FoldAccInit = {Limit, Skip, undefined, []},
-        {ok, {_, _, Resp, _}} = couch_view:fold_reduce(View, Dir, {StartKey, StartDocId},
-            {EndKey, EndDocId}, GroupRowsFun, RespFun, FoldAccInit),
+        {ok, {_, _, Resp, _}} = couch_view:fold_reduce(View,
+                RespFun, FoldAccInit, [{key_group_fun, GroupRowsFun} |
+                make_key_options(QueryArgs)]),
         finish_reduce_fold(Req, Resp)
     end);
 
@@ -208,20 +163,19 @@ output_reduce_view(Req, Db, View, Group, QueryArgs, Keys) ->
     #view_query_args{
         limit = Limit,
         skip = Skip,
-        direction = Dir,
-        start_docid = StartDocId,
-        end_docid = EndDocId,
         group_level = GroupLevel
     } = QueryArgs,
-    CurrentEtag = view_group_etag(Group, Db),
+    CurrentEtag = view_group_etag(Group, Db, Keys),
     couch_httpd:etag_respond(Req, CurrentEtag, fun() ->
         {ok, GroupRowsFun, RespFun} = make_reduce_fold_funs(Req, GroupLevel, QueryArgs, CurrentEtag, #reduce_fold_helper_funs{}),
         {Resp, _RedAcc3} = lists:foldl(
             fun(Key, {Resp, RedAcc}) ->
                 % run the reduce once for each key in keys, with limit etc reapplied for each key
                 FoldAccInit = {Limit, Skip, Resp, RedAcc},
-                {_, {_, _, Resp2, RedAcc2}} = couch_view:fold_reduce(View, Dir, {Key, StartDocId},
-                    {Key, EndDocId}, GroupRowsFun, RespFun, FoldAccInit),
+                {_, {_, _, Resp2, RedAcc2}} = couch_view:fold_reduce(View,
+                        RespFun, FoldAccInit, [{key_group_fun, GroupRowsFun} |
+                        make_key_options(QueryArgs#view_query_args{
+                            start_key=Key, end_key=Key})]),
                 % Switch to comma
                 {Resp2, RedAcc2}
             end,
@@ -229,8 +183,8 @@ output_reduce_view(Req, Db, View, Group, QueryArgs, Keys) ->
         finish_reduce_fold(Req, Resp)
     end).
 
-reverse_key_default(nil) -> {};
-reverse_key_default({}) -> nil;
+reverse_key_default(?MIN_STR) -> ?MAX_STR;
+reverse_key_default(?MAX_STR) -> ?MIN_STR;
 reverse_key_default(Key) -> Key.
 
 get_stale_type(Req) ->
@@ -239,16 +193,42 @@ get_stale_type(Req) ->
 get_reduce_type(Req) ->
     list_to_atom(couch_httpd:qs_value(Req, "reduce", "true")).
 
+load_view(Req, Db, {ViewDesignId, ViewName}, Keys) ->
+    Stale = couch_httpd_view:get_stale_type(Req),
+    Reduce = couch_httpd_view:get_reduce_type(Req),
+    case couch_view:get_map_view(Db, ViewDesignId, ViewName, Stale) of
+    {ok, View, Group} ->
+        QueryArgs = couch_httpd_view:parse_view_params(Req, Keys, map),
+        {map, View, Group, QueryArgs};
+    {not_found, _Reason} ->
+        case couch_view:get_reduce_view(Db, ViewDesignId, ViewName, Stale) of
+        {ok, ReduceView, Group} ->
+            case Reduce of
+            false ->
+                QueryArgs = couch_httpd_view:parse_view_params(Req, Keys, map_red),
+                MapView = couch_view:extract_map_view(ReduceView),
+                {map, MapView, Group, QueryArgs};
+            _ ->
+                QueryArgs = couch_httpd_view:parse_view_params(Req, Keys, reduce),                
+                {reduce, ReduceView, Group, QueryArgs}
+            end;
+        {not_found, Reason} ->
+            throw({not_found, Reason})
+        end
+    end.
+
+% query_parse_error could be removed
+% we wouldn't need to pass the view type, it'd just parse params.
+% I'm not sure what to do about the error handling, but
+% it might simplify things to have a parse_view_params function
+% that doesn't throw().
 parse_view_params(Req, Keys, ViewType) ->
     QueryList = couch_httpd:qs(Req),
     QueryParams =
     lists:foldl(fun({K, V}, Acc) ->
             parse_view_param(K, V) ++ Acc
         end, [], QueryList),
-    IsMultiGet = case Keys of
-        nil -> false;
-        _ -> true
-    end,
+    IsMultiGet = (Keys =/= nil),
     Args = #view_query_args{
         view_type=ViewType,
         multi_get=IsMultiGet
@@ -264,6 +244,7 @@ parse_view_params(Req, Keys, ViewType) ->
         {reduce, _, false} ->
             QueryArgs;
         {reduce, _, _} ->
+            % we can simplify code if we just drop this error message.
             Msg = <<"Multi-key fetchs for reduce "
                     "view must include `group=true`">>,
             throw({query_parse_error, Msg});
@@ -323,7 +304,7 @@ validate_view_query(start_key, Value, Args) ->
     case Args#view_query_args.multi_get of
         true ->
             Msg = <<"Query parameter `start_key` is "
-                    "not compatiible with multi-get">>,
+                    "not compatible with multi-get">>,
             throw({query_parse_error, Msg});
         _ ->
             Args#view_query_args{start_key=Value}
@@ -333,8 +314,8 @@ validate_view_query(start_docid, Value, Args) ->
 validate_view_query(end_key, Value, Args) ->
     case Args#view_query_args.multi_get of
         true->
-            Msg = <<"Query paramter `end_key` is "
-                    "not compatibile with multi-get">>,
+            Msg = <<"Query parameter `end_key` is "
+                    "not compatible with multi-get">>,
             throw({query_parse_error, Msg});
         _ ->
             Args#view_query_args{end_key=Value}
@@ -353,12 +334,8 @@ validate_view_query(descending, true, Args) ->
         fwd ->
             Args#view_query_args{
                 direction = rev,
-                start_key =
-                    reverse_key_default(Args#view_query_args.start_key),
                 start_docid =
                     reverse_key_default(Args#view_query_args.start_docid),
-                end_key =
-                    reverse_key_default(Args#view_query_args.end_key),
                 end_docid =
                     reverse_key_default(Args#view_query_args.end_docid)
             }
@@ -389,7 +366,7 @@ validate_view_query(reduce, _, Args) ->
 validate_view_query(include_docs, true, Args) ->
     case Args#view_query_args.view_type of
         reduce ->
-            Msg = <<"Query paramter `include_docs` "
+            Msg = <<"Query parameter `include_docs` "
                     "is invalid for reduce views.">>,
             throw({query_parse_error, Msg});
         _ ->
@@ -401,58 +378,37 @@ validate_view_query(extra, _Value, Args) ->
     Args.
 
 make_view_fold_fun(Req, QueryArgs, Etag, Db, TotalViewCount, HelperFuns) ->
-    #view_query_args{
-        end_key = EndKey,
-        end_docid = EndDocId,
-        inclusive_end = InclusiveEnd,
-        direction = Dir
-    } = QueryArgs,
-
     #view_fold_helper_funs{
-        passed_end = PassedEndFun,
         start_response = StartRespFun,
         send_row = SendRowFun,
         reduce_count = ReduceCountFun
-    } = apply_default_helper_funs(HelperFuns,
-        {Dir, EndKey, EndDocId, InclusiveEnd}),
+    } = apply_default_helper_funs(HelperFuns),
 
     #view_query_args{
         include_docs = IncludeDocs
     } = QueryArgs,
-
-    fun({{Key, DocId}, Value}, OffsetReds, {AccLimit, AccSkip, Resp, RowFunAcc,
-                                            OffsetAcc}) ->
-        PassedEnd = PassedEndFun(Key, DocId),
-        case {PassedEnd, AccLimit, AccSkip, Resp} of
-        {true, _, _, _} ->
-            % The stop key has been passed, stop looping.
-            % We may need offset so calcluate it here.
-            % Checking Resp is an optimization that tells
-            % us its already been calculated (and sent).
-            NewOffset = case Resp of
-                undefined -> ReduceCountFun(OffsetReds);
-                _ -> nil
-            end,
-            {stop, {AccLimit, AccSkip, Resp, RowFunAcc, NewOffset}};
-        {_, 0, _, _} ->
+    
+    fun({{Key, DocId}, Value}, OffsetReds, {AccLimit, AccSkip, Resp, RowFunAcc}) ->
+        case {AccLimit, AccSkip, Resp} of
+        {0, _, _} ->
             % we've done "limit" rows, stop foldling
-            {stop, {0, 0, Resp, RowFunAcc, OffsetAcc}};
-        {_, _, AccSkip, _} when AccSkip > 0 ->
+            {stop, {0, 0, Resp, RowFunAcc}};
+        {_, AccSkip, _} when AccSkip > 0 ->
             % just keep skipping
-            {ok, {AccLimit, AccSkip - 1, Resp, RowFunAcc, OffsetAcc}};
-        {_, _, _, undefined} ->
+            {ok, {AccLimit, AccSkip - 1, Resp, RowFunAcc}};
+        {_, _, undefined} ->
             % rendering the first row, first we start the response
             Offset = ReduceCountFun(OffsetReds),
             {ok, Resp2, RowFunAcc0} = StartRespFun(Req, Etag,
                 TotalViewCount, Offset, RowFunAcc),
             {Go, RowFunAcc2} = SendRowFun(Resp2, Db, {{Key, DocId}, Value},
                 IncludeDocs, RowFunAcc0),
-            {Go, {AccLimit - 1, 0, Resp2, RowFunAcc2, Offset}};
-        {_, AccLimit, _, Resp} when (AccLimit > 0) ->
+            {Go, {AccLimit - 1, 0, Resp2, RowFunAcc2}};
+        {AccLimit, _, Resp} when (AccLimit > 0) ->
             % rendering all other rows
             {Go, RowFunAcc2} = SendRowFun(Resp, Db, {{Key, DocId}, Value},
                 IncludeDocs, RowFunAcc),
-            {Go, {AccLimit - 1, 0, Resp, RowFunAcc2, OffsetAcc}}
+            {Go, {AccLimit - 1, 0, Resp, RowFunAcc2}}
         end
     end.
 
@@ -515,14 +471,9 @@ make_reduce_fold_funs(Req, GroupLevel, _QueryArgs, Etag, HelperFuns) ->
     {ok, GroupRowsFun, RespFun}.
 
 apply_default_helper_funs(#view_fold_helper_funs{
-    passed_end = PassedEnd,
     start_response = StartResp,
     send_row = SendRow
-}=Helpers, {Dir, EndKey, EndDocId, InclusiveEnd}) ->
-    PassedEnd2 = case PassedEnd of
-    undefined -> make_passed_end_fun(Dir, EndKey, EndDocId, InclusiveEnd);
-    _ -> PassedEnd
-    end,
+}=Helpers) ->
 
     StartResp2 = case StartResp of
     undefined -> fun json_view_start_resp/5;
@@ -535,10 +486,10 @@ apply_default_helper_funs(#view_fold_helper_funs{
     end,
 
     Helpers#view_fold_helper_funs{
-        passed_end = PassedEnd2,
         start_response = StartResp2,
         send_row = SendRow2
-    }.
+    };
+
 
 apply_default_helper_funs(#reduce_fold_helper_funs{
     start_response = StartResp,
@@ -559,36 +510,34 @@ apply_default_helper_funs(#reduce_fold_helper_funs{
         send_row = SendRow2
     }.
 
-make_passed_end_fun(fwd, EndKey, EndDocId, InclusiveEnd) ->
-    case InclusiveEnd of
-    true ->
-        fun(ViewKey, ViewId) ->
-            couch_view:less_json([EndKey, EndDocId], [ViewKey, ViewId])
-        end;
-    false ->
-        fun
-            (ViewKey, _ViewId) when ViewKey == EndKey ->
-                true;
-            (ViewKey, ViewId) ->
-                couch_view:less_json([EndKey, EndDocId], [ViewKey, ViewId])
-        end
-    end;
+make_key_options(#view_query_args{direction = Dir}=QueryArgs) ->
+     [{dir,Dir} | make_start_key_option(QueryArgs) ++ 
+            make_end_key_option(QueryArgs)].
 
-make_passed_end_fun(rev, EndKey, EndDocId, InclusiveEnd) ->
-    case InclusiveEnd of
+make_start_key_option(
+        #view_query_args{
+            start_key = StartKey,
+            start_docid = StartDocId}) ->
+    if StartKey == undefined ->
+        [];
     true ->
-        fun(ViewKey, ViewId) ->
-            couch_view:less_json([ViewKey, ViewId], [EndKey, EndDocId])
-        end;
-    false->
-        fun
-            (ViewKey, _ViewId) when ViewKey == EndKey ->
-                true;
-            (ViewKey, ViewId) ->
-                couch_view:less_json([ViewKey, ViewId], [EndKey, EndDocId])
-        end
+        [{start_key, {StartKey, StartDocId}}]
     end.
 
+make_end_key_option(#view_query_args{end_key = undefined}) ->
+    [];
+make_end_key_option(
+        #view_query_args{end_key = EndKey,
+            end_docid = EndDocId,
+            inclusive_end = true}) ->
+    [{end_key, {EndKey, EndDocId}}];
+make_end_key_option(
+        #view_query_args{
+            end_key = EndKey,
+            end_docid = EndDocId,
+            inclusive_end = false}) ->
+    [{end_key_gt, {EndKey,reverse_key_default(EndDocId)}}].
+
 json_view_start_resp(Req, Etag, TotalViewCount, Offset, _Acc) ->
     {ok, Resp} = start_json_response(Req, 200, [{"Etag", Etag}]),
     BeginBody = io_lib:format("{\"total_rows\":~w,\"offset\":~w,\"rows\":[\r\n",
@@ -630,47 +579,42 @@ view_row_obj(Db, {{Key, DocId}, {Props}}, true) ->
     Rev0 ->
         couch_doc:parse_rev(Rev0)
     end,
-    view_row_with_doc(Db, {{Key, DocId}, {Props}}, Rev);
+    IncludeId = proplists:get_value(<<"_id">>, Props, DocId),
+    view_row_with_doc(Db, {{Key, DocId}, {Props}}, {IncludeId, Rev});
 view_row_obj(Db, {{Key, DocId}, Value}, true) ->
-    view_row_with_doc(Db, {{Key, DocId}, Value}, nil);
+    view_row_with_doc(Db, {{Key, DocId}, Value}, {DocId, nil});
 % the normal case for rendering a view row
 view_row_obj(_Db, {{Key, DocId}, Value}, _IncludeDocs) ->
     {[{id, DocId}, {key, Key}, {value, Value}]}.
 
-view_row_with_doc(Db, {{Key, DocId}, Value}, Rev) ->
+view_row_with_doc(Db, {{Key, DocId}, Value}, IdRev) ->
+    {[{id, DocId}, {key, Key}, {value, Value}] ++ doc_member(Db, IdRev)}.
+
+doc_member(Db, {DocId, Rev}) ->
     ?LOG_DEBUG("Include Doc: ~p ~p", [DocId, Rev]),
     case (catch couch_httpd_db:couch_doc_open(Db, DocId, Rev, [])) of
-      {{not_found, missing}, _RevId} ->
-          {[{id, DocId}, {key, Key}, {value, Value}, {error, missing}]};
-      {not_found, missing} ->
-          {[{id, DocId}, {key, Key}, {value, Value}, {error, missing}]};
-      {not_found, deleted} ->
-          {[{id, DocId}, {key, Key}, {value, Value}]};
-      Doc ->
-        JsonDoc = couch_doc:to_json_obj(Doc, []),
-        {[{id, DocId}, {key, Key}, {value, Value}, {doc, JsonDoc}]}
+        #doc{} = Doc ->
+            JsonDoc = couch_doc:to_json_obj(Doc, []),
+            [{doc, JsonDoc}];
+        _Else ->
+            [{doc, null}]
     end.
+    
 
-finish_view_fold(Req, TotalRows, FoldResult) ->
+finish_view_fold(Req, TotalRows, Offset, FoldResult) ->
     case FoldResult of
-    {ok, {_, _, undefined, _, Offset}} ->
-        % nothing found in the view, nothing has been returned
+    {_, _, undefined, _} ->
+        % nothing found in the view or keys, nothing has been returned
         % send empty view
-        NewOffset = case Offset of
-            nil -> TotalRows;
-            _ -> Offset
-        end,
         send_json(Req, 200, {[
             {total_rows, TotalRows},
-            {offset, NewOffset},
+            {offset, Offset},
             {rows, []}
         ]});
-    {ok, {_, _, Resp, _, _}} ->
+    {_, _, Resp, _} ->
         % end the view
         send_chunk(Resp, "\r\n]}"),
-        end_json_response(Resp);
-    Error ->
-        throw(Error)
+        end_json_response(Resp)
     end.
 
 finish_reduce_fold(Req, Resp) ->
@@ -684,11 +628,14 @@ finish_reduce_fold(Req, Resp) ->
         end_json_response(Resp)
     end.
 
-parse_bool_param("true") -> true;
-parse_bool_param("false") -> false;
 parse_bool_param(Val) ->
-    Msg = io_lib:format("Invalid value for boolean paramter: ~p", [Val]),
-    throw({query_parse_error, ?l2b(Msg)}).
+    case string:to_lower(Val) of
+        "true" -> true;
+        "false" -> false;
+        _ ->
+            Msg = io_lib:format("Invalid boolean parameter: ~p", [Val]),
+            throw({query_parse_error, ?l2b(Msg)})
+    end.
 
 parse_int_param(Val) ->
     case (catch list_to_integer(Val)) of
