@@ -28,6 +28,7 @@
     source,
     target,
     continuous,
+    create_target,
     init_args,
     checkpoint_scheduled = nil,
 
@@ -44,7 +45,8 @@
     complete = false,
     committed_seq = 0,
 
-    stats = nil
+    stats = nil,
+    doc_ids = nil
 }).
 
 %% convenience function to do a simple replication from the shell
@@ -66,13 +68,24 @@ replicate({Props}=PostBody, UserCtx) ->
         [?MODULE]
     },
 
-    Server = start_replication_server(Replicator),
-
-    case proplists:get_value(<<"continuous">>, Props, false) of
+    case proplists:get_value(<<"cancel">>, Props, false) of
     true ->
-        {ok, {continuous, ?l2b(BaseId)}};
+ case supervisor:terminate_child(couch_rep_sup, BaseId ++ Extension) of
+        {error, not_found} ->
+     {error, not_found};
+        ok ->
+     ok = supervisor:delete_child(couch_rep_sup, BaseId ++ Extension),
+            {ok, {cancelled, ?l2b(BaseId)}}
+ end;
     false ->
-        get_result(Server, PostBody, UserCtx)
+        Server = start_replication_server(Replicator),
+
+        case proplists:get_value(<<"continuous">>, Props, false) of
+        true ->
+            {ok, {continuous, ?l2b(BaseId)}};
+        false ->
+            get_result(Server, PostBody, UserCtx)
+        end
     end.
 
 checkpoint(Server) ->
@@ -101,25 +114,49 @@ do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
     SourceProps = proplists:get_value(<<"source">>, PostProps),
     TargetProps = proplists:get_value(<<"target">>, PostProps),
 
+    DocIds = proplists:get_value(<<"doc_ids">>, PostProps, nil),
     Continuous = proplists:get_value(<<"continuous">>, PostProps, false),
+    CreateTarget = proplists:get_value(<<"create_target">>, PostProps, false),
 
     Source = open_db(SourceProps, UserCtx),
-    Target = open_db(TargetProps, UserCtx),
-
-    SourceLog = open_replication_log(Source, RepId),
-    TargetLog = open_replication_log(Target, RepId),
+    Target = open_db(TargetProps, UserCtx, CreateTarget),
 
     SourceInfo = dbinfo(Source),
     TargetInfo = dbinfo(Target),
-    
-    {StartSeq, History} = compare_replication_logs(SourceLog, TargetLog),
 
-    {ok, ChangesFeed} =
-    couch_rep_changes_feed:start_link(self(), Source, StartSeq, PostProps),
-    {ok, MissingRevs} =
-    couch_rep_missing_revs:start_link(self(), Target, ChangesFeed, PostProps),
-    {ok, Reader} =
-    couch_rep_reader:start_link(self(), Source, MissingRevs, PostProps),
+    case DocIds of
+    List when is_list(List) ->
+        % Fast replication using only a list of doc IDs to replicate.
+        % Replication sessions, checkpoints and logs are not created
+        % since the update sequence number of the source DB is not used
+        % for determining which documents are copied into the target DB.
+        SourceLog = nil,
+        TargetLog = nil,
+
+        StartSeq = nil,
+        History = nil,
+
+        ChangesFeed = nil,
+        MissingRevs = nil,
+
+        {ok, Reader} =
+        couch_rep_reader:start_link(self(), Source, DocIds, PostProps);
+
+    _ ->
+        % Replication using the _changes API (DB sequence update numbers).
+        SourceLog = open_replication_log(Source, RepId),
+        TargetLog = open_replication_log(Target, RepId),
+    
+        {StartSeq, History} = compare_replication_logs(SourceLog, TargetLog),
+
+        {ok, ChangesFeed} =
+        couch_rep_changes_feed:start_link(self(), Source, StartSeq, PostProps),
+        {ok, MissingRevs} =
+        couch_rep_missing_revs:start_link(self(), Target, ChangesFeed, PostProps),
+        {ok, Reader} =
+        couch_rep_reader:start_link(self(), Source, MissingRevs, PostProps)
+    end,
+
     {ok, Writer} =
     couch_rep_writer:start_link(self(), Target, Reader, PostProps),
 
@@ -143,6 +180,7 @@ do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
         source = Source,
         target = Target,
         continuous = Continuous,
+        create_target = CreateTarget,
         init_args = InitArgs,
         stats = Stats,
         checkpoint_scheduled = nil,
@@ -153,7 +191,8 @@ do_init([RepId, {PostProps}, UserCtx] = InitArgs) ->
         target_log = TargetLog,
         rep_starttime = httpd_util:rfc1123_date(),
         src_starttime = proplists:get_value(instance_start_time, SourceInfo),
-        tgt_starttime = proplists:get_value(instance_start_time, TargetInfo)
+        tgt_starttime = proplists:get_value(instance_start_time, TargetInfo),
+        doc_ids = DocIds
     },
     {ok, State}.
 
@@ -281,7 +320,7 @@ compare_replication_logs(SrcDoc, TgtDoc) ->
         compare_rep_history(SourceHistory, TargetHistory)
     end.
 
-compare_rep_history(S, T) when length(S) =:= 0 orelse length(T) =:= 0 ->
+compare_rep_history(S, T) when S =:= [] orelse T =:= [] ->
     ?LOG_INFO("no common ancestry -- performing full replication", []),
     {0, []};
 compare_rep_history([{S}|SourceRest], [{T}|TargetRest]=Target) ->
@@ -305,15 +344,15 @@ compare_rep_history([{S}|SourceRest], [{T}|TargetRest]=Target) ->
         end
     end.
 
-close_db(#http_db{})->
+close_db(#http_db{}) ->
     ok;
-close_db(Db)->
+close_db(Db) ->
     couch_db:close(Db).
 
-dbname(#http_db{} = Db) ->
-    Db#http_db.url;
-dbname(Db) ->
-    Db#db.name.
+dbname(#http_db{url = Url}) ->
+    Url;
+dbname(#db{name = Name}) ->
+    Name.
 
 dbinfo(#http_db{} = Db) ->
     {DbProps} = couch_rep_httpc:request(Db),
@@ -322,20 +361,34 @@ dbinfo(Db) ->
     {ok, Info} = couch_db:get_db_info(Db),
     Info.
 
+do_terminate(#state{doc_ids=DocIds} = State) when is_list(DocIds) ->
+    #state{
+        listeners = Listeners,
+        rep_starttime = ReplicationStartTime,
+        stats = Stats
+    } = State,
+
+    RepByDocsJson = {[
+        {<<"start_time">>, ?l2b(ReplicationStartTime)},
+        {<<"end_time">>, ?l2b(httpd_util:rfc1123_date())},
+        {<<"docs_read">>, ets:lookup_element(Stats, docs_read, 2)},
+        {<<"docs_written">>, ets:lookup_element(Stats, docs_written, 2)},
+        {<<"doc_write_failures">>,
+            ets:lookup_element(Stats, doc_write_failures, 2)}
+    ]},
+
+    terminate_cleanup(State),
+    [gen_server:reply(L, {ok, RepByDocsJson}) || L <- lists:reverse(Listeners)];
+
 do_terminate(State) ->
     #state{
         checkpoint_history = CheckpointHistory,
         committed_seq = NewSeq,
         listeners = Listeners,
         source = Source,
-        target = Target,
         continuous = Continuous,
-        stats = Stats,
         source_log = #doc{body={OldHistory}}
     } = State,
-    couch_task_status:update("Finishing"),
-    ets:delete(Stats),
-    close_db(Target),
     
     NewRepHistory = case CheckpointHistory of
     nil ->
@@ -363,7 +416,13 @@ do_terminate(State) ->
         false ->
             [gen_server:reply(R, retry) || R <- OtherListeners]
     end,
-    close_db(Source).
+    terminate_cleanup(State).
+
+terminate_cleanup(#state{source=Source, target=Target, stats=Stats}) ->
+    couch_task_status:update("Finishing"),
+    close_db(Target),
+    close_db(Source),
+    ets:delete(Stats).
 
 has_session_id(_SessionId, []) ->
     false;
@@ -375,20 +434,38 @@ has_session_id(SessionId, [{Props} | Rest]) ->
         has_session_id(SessionId, Rest)
     end.
 
+maybe_append_options(Options, Props) ->
+    lists:foldl(fun(Option, Acc) ->
+        Acc ++ 
+        case proplists:get_value(Option, Props, false) of
+        true ->
+            "+" ++ ?b2l(Option);
+        false ->
+            ""
+        end
+    end, [], Options).
+
 make_replication_id({Props}, UserCtx) ->
     %% funky algorithm to preserve backwards compatibility
     {ok, HostName} = inet:gethostname(),
     % Port = mochiweb_socket_server:get(couch_httpd, port),
     Src = get_rep_endpoint(UserCtx, proplists:get_value(<<"source">>, Props)),
     Tgt = get_rep_endpoint(UserCtx, proplists:get_value(<<"target">>, Props)),    
-    Base = couch_util:to_hex(erlang:md5(term_to_binary([HostName, Src, Tgt]))),
-    Extension = case proplists:get_value(<<"continuous">>, Props, false) of
-    true ->
-        "+continuous";
-    false ->
-        ""
-    end,
-    {Base, Extension}.
+    Base = [HostName, Src, Tgt] ++
+        case proplists:get_value(<<"filter">>, Props) of
+        undefined ->
+            case proplists:get_value(<<"doc_ids">>, Props) of
+            undefined ->
+                [];
+            DocIds ->
+                [DocIds]
+            end;
+        Filter ->
+            [Filter, proplists:get_value(<<"query_params">>, Props, {[]})]
+        end,
+    Extension = maybe_append_options(
+        [<<"continuous">>, <<"create_target">>], Props),
+    {couch_util:to_hex(erlang:md5(term_to_binary(Base))), Extension}.
 
 maybe_add_trailing_slash(Url) ->
     re:replace(Url, "[^/]$", "&/", [{return, list}]).
@@ -432,7 +509,10 @@ open_replication_log(Db, RepId) ->
         #doc{id=DocId}
     end.
 
-open_db({Props}, _UserCtx) ->
+open_db(Props, UserCtx) ->
+    open_db(Props, UserCtx, false).
+
+open_db({Props}, _UserCtx, CreateTarget) ->
     Url = maybe_add_trailing_slash(proplists:get_value(<<"url">>, Props)),
     {AuthProps} = proplists:get_value(<<"auth">>, Props, {[]}),
     {BinHeaders} = proplists:get_value(<<"headers">>, Props, {[]}),
@@ -443,12 +523,19 @@ open_db({Props}, _UserCtx) ->
         auth = AuthProps,
         headers = lists:ukeymerge(1, Headers, DefaultHeaders)
     },
-    couch_rep_httpc:db_exists(Db);
-open_db(<<"http://",_/binary>>=Url, _) ->
-    open_db({[{<<"url">>,Url}]}, []);
-open_db(<<"https://",_/binary>>=Url, _) ->
-    open_db({[{<<"url">>,Url}]}, []);
-open_db(<<DbName/binary>>, UserCtx) ->
+    couch_rep_httpc:db_exists(Db, CreateTarget);
+open_db(<<"http://",_/binary>>=Url, _, CreateTarget) ->
+    open_db({[{<<"url">>,Url}]}, [], CreateTarget);
+open_db(<<"https://",_/binary>>=Url, _, CreateTarget) ->
+    open_db({[{<<"url">>,Url}]}, [], CreateTarget);
+open_db(<<DbName/binary>>, UserCtx, CreateTarget) ->
+    case CreateTarget of
+    true -> 
+        ok = couch_httpd:verify_is_server_admin(UserCtx),
+        couch_server:create(DbName, [{user_ctx, UserCtx}]);
+    false -> ok
+    end,
+
     case couch_db:open(DbName, [{user_ctx, UserCtx}]) of
     {ok, Db} ->
         couch_db:monitor(Db),
@@ -482,53 +569,64 @@ do_checkpoint(State) ->
         tgt_starttime = TgtInstanceStartTime,
         stats = Stats
     } = State,
-    ?LOG_INFO("recording a checkpoint at source update_seq ~p", [NewSeqNum]),
-    RecordSeqNum = case commit_to_both(Source, Target, NewSeqNum) of
+    case commit_to_both(Source, Target, NewSeqNum) of
     {SrcInstanceStartTime, TgtInstanceStartTime} ->
-        NewSeqNum;
-    _Else ->
-        ?LOG_INFO("A server has restarted sinced replication start. "
-            "Not recording the new sequence number to ensure the "
-            "replication is redone and documents reexamined.", []),
-        StartSeqNum
-    end,
-    SessionId = couch_util:new_uuid(),
-    NewHistoryEntry = {[
-        {<<"session_id">>, SessionId},
-        {<<"start_time">>, list_to_binary(ReplicationStartTime)},
-        {<<"end_time">>, list_to_binary(httpd_util:rfc1123_date())},
-        {<<"start_last_seq">>, StartSeqNum},
-        {<<"end_last_seq">>, NewSeqNum},
-        {<<"recorded_seq">>, RecordSeqNum},
-        {<<"missing_checked">>, ets:lookup_element(Stats, total_revs, 2)},
-        {<<"missing_found">>, ets:lookup_element(Stats, missing_revs, 2)},
-        {<<"docs_read">>, ets:lookup_element(Stats, docs_read, 2)},
-        {<<"docs_written">>, ets:lookup_element(Stats, docs_written, 2)},
-        {<<"doc_write_failures">>, 
-            ets:lookup_element(Stats, doc_write_failures, 2)}
-    ]},
-    % limit history to 50 entries
-    NewRepHistory = {[
-        {<<"session_id">>, SessionId},
-        {<<"source_last_seq">>, RecordSeqNum},
-        {<<"history">>, lists:sublist([NewHistoryEntry | OldHistory], 50)}
-    ]},
+        ?LOG_INFO("recording a checkpoint for ~s -> ~s at source update_seq ~p",
+            [dbname(Source), dbname(Target), NewSeqNum]),
+        SessionId = couch_uuids:random(),
+        NewHistoryEntry = {[
+            {<<"session_id">>, SessionId},
+            {<<"start_time">>, list_to_binary(ReplicationStartTime)},
+            {<<"end_time">>, list_to_binary(httpd_util:rfc1123_date())},
+            {<<"start_last_seq">>, StartSeqNum},
+            {<<"end_last_seq">>, NewSeqNum},
+            {<<"recorded_seq">>, NewSeqNum},
+            {<<"missing_checked">>, ets:lookup_element(Stats, total_revs, 2)},
+            {<<"missing_found">>, ets:lookup_element(Stats, missing_revs, 2)},
+            {<<"docs_read">>, ets:lookup_element(Stats, docs_read, 2)},
+            {<<"docs_written">>, ets:lookup_element(Stats, docs_written, 2)},
+            {<<"doc_write_failures">>, 
+                ets:lookup_element(Stats, doc_write_failures, 2)}
+        ]},
+        % limit history to 50 entries
+        NewRepHistory = {[
+            {<<"session_id">>, SessionId},
+            {<<"source_last_seq">>, NewSeqNum},
+            {<<"history">>, lists:sublist([NewHistoryEntry | OldHistory], 50)}
+        ]},
 
-    try
-    {SrcRevPos,SrcRevId} =
-        update_local_doc(Source, SourceLog#doc{body=NewRepHistory}),
-    {TgtRevPos,TgtRevId} =
-        update_local_doc(Target, TargetLog#doc{body=NewRepHistory}),
-    State#state{
-        checkpoint_scheduled = nil,
-        checkpoint_history = NewRepHistory,
-        source_log = SourceLog#doc{revs={SrcRevPos, [SrcRevId]}},
-        target_log = TargetLog#doc{revs={TgtRevPos, [TgtRevId]}}
-    }
-    catch throw:conflict ->
-    ?LOG_ERROR("checkpoint failure: conflict (are you replicating to yourself?)",
-        []),
-    State
+        try
+        {SrcRevPos,SrcRevId} =
+            update_local_doc(Source, SourceLog#doc{body=NewRepHistory}),
+        {TgtRevPos,TgtRevId} =
+            update_local_doc(Target, TargetLog#doc{body=NewRepHistory}),
+        State#state{
+            checkpoint_scheduled = nil,
+            checkpoint_history = NewRepHistory,
+            source_log = SourceLog#doc{revs={SrcRevPos, [SrcRevId]}},
+            target_log = TargetLog#doc{revs={TgtRevPos, [TgtRevId]}}
+        }
+        catch throw:conflict ->
+        ?LOG_ERROR("checkpoint failure: conflict (are you replicating to "
+            "yourself?)", []),
+        State
+        end;
+    _Else ->
+        ?LOG_INFO("rebooting ~s -> ~s from last known replication checkpoint",
+            [dbname(Source), dbname(Target)]),
+        #state{
+            changes_feed = CF,
+            missing_revs = MR,
+            reader = Reader,
+            writer = Writer
+        } = State,
+        Pids = [Writer, Reader, MR, CF],
+        [unlink(Pid) || Pid <- Pids],
+        [exit(Pid, shutdown) || Pid <- Pids],
+        close_db(Target),
+        close_db(Source),
+        {ok, NewState} = init(State#state.init_args),
+        NewState#state{listeners=State#state.listeners}
     end.
 
 commit_to_both(Source, Target, RequiredSeq) ->
@@ -559,18 +657,18 @@ ensure_full_commit(#http_db{} = Target) ->
     true = proplists:get_value(<<"ok">>, ResultProps),
     proplists:get_value(<<"instance_start_time">>, ResultProps);
 ensure_full_commit(Target) ->
-    {ok, NewDb} = couch_db:open(Target#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Target#db.name, []),
     UpdateSeq = couch_db:get_update_seq(Target),
     CommitSeq = couch_db:get_committed_update_seq(NewDb),
     InstanceStartTime = NewDb#db.instance_start_time,
     couch_db:close(NewDb),
     if UpdateSeq > CommitSeq ->
-        ?LOG_DEBUG("replication needs a full commit: update ~p commit ~p",
+        ?LOG_DEBUG("target needs a full commit: update ~p commit ~p",
             [UpdateSeq, CommitSeq]),
         {ok, DbStartTime} = couch_db:ensure_full_commit(Target),
         DbStartTime;
     true ->
-        ?LOG_DEBUG("replication doesn't need a full commit", []),
+        ?LOG_DEBUG("target doesn't need a full commit", []),
         InstanceStartTime
     end.
 
@@ -587,14 +685,17 @@ ensure_full_commit(#http_db{} = Source, RequiredSeq) ->
         proplists:get_value(<<"instance_start_time">>, ResultProps);
     undefined -> nil end;
 ensure_full_commit(Source, RequiredSeq) ->
-    {ok, NewDb} = couch_db:open(Source#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Source#db.name, []),
     CommitSeq = couch_db:get_committed_update_seq(NewDb),
     InstanceStartTime = NewDb#db.instance_start_time,
     couch_db:close(NewDb),
     if RequiredSeq > CommitSeq ->
+        ?LOG_DEBUG("source needs a full commit: required ~p committed ~p",
+            [RequiredSeq, CommitSeq]),
         {ok, DbStartTime} = couch_db:ensure_full_commit(Source),
         DbStartTime;
     true ->
+        ?LOG_DEBUG("source doesn't need a full commit", []),
         InstanceStartTime
     end.
 
@@ -615,7 +716,7 @@ update_local_doc(Db, Doc) ->
 up_to_date(#http_db{}, _Seq) ->
     true;
 up_to_date(Source, Seq) ->
-    {ok, NewDb} = couch_db:open(Source#db.name, []),
+    {ok, NewDb} = couch_db:open_int(Source#db.name, []),
     T = NewDb#db.update_seq == Seq,
     couch_db:close(NewDb),
     T.
