@@ -12,14 +12,17 @@
 
 -module(couch_util).
 
--export([start_driver/1,terminate_linked/1]).
+-export([priv_dir/0, start_driver/1, normpath/1, terminate_linked/1]).
 -export([should_flush/0, should_flush/1, to_existing_atom/1]).
--export([new_uuid/0, rand32/0, implode/2, collate/2, collate/3]).
+-export([rand32/0, implode/2, collate/2, collate/3]).
 -export([abs_pathname/1,abs_pathname/2, trim/1, ascii_lower/1]).
 -export([encodeBase64/1, decodeBase64/1, encodeBase64Url/1, decodeBase64Url/1,
     to_hex/1,parse_term/1, dict_find/3]).
 -export([file_read_size/1, get_nested_json_value/2, json_user_ctx/1]).
--export([to_binary/1, to_list/1, url_encode/1]).
+-export([to_binary/1, to_integer/1, to_list/1, url_encode/1]).
+-export([json_encode/1, json_decode/1]).
+-export([verify/2]).
+-export([compressible_att_type/1]).
 
 -include("couch_db.hrl").
 -include_lib("kernel/include/file.hrl").
@@ -27,23 +30,46 @@
 % arbitrarily chosen amount of memory to use before flushing to disk
 -define(FLUSH_MAX_MEM, 10000000).
 
+priv_dir() ->
+    case code:priv_dir(couch) of
+        {error, bad_name} ->
+            % small hack, in dev mode "app" is couchdb. Fixing requires
+            % renaming src/couch to src/couch. Not really worth the hassle.
+            % -Damien
+            code:priv_dir(couchdb);
+        Dir -> Dir
+    end.
+
 start_driver(LibDir) ->
-    case erl_ddll:load_driver(LibDir, "couch_erl_driver") of
+    case erl_ddll:load_driver(LibDir, "couch_icu_driver") of
     ok ->
         ok;
     {error, already_loaded} ->
-        ok = erl_ddll:reload_driver(LibDir, "couch_erl_driver");
+        ok = erl_ddll:reload_driver(LibDir, "couch_icu_driver");
     {error, Error} ->
         exit(erl_ddll:format_error(Error))
     end.
 
+% Normalize a pathname by removing .. and . components.
+normpath(Path) ->
+    normparts(filename:split(Path), []).
+
+normparts([], Acc) ->
+    filename:join(lists:reverse(Acc));
+normparts([".." | RestParts], [_Drop | RestAcc]) ->
+    normparts(RestParts, RestAcc);
+normparts(["." | RestParts], Acc) ->
+    normparts(RestParts, Acc);
+normparts([Part | RestParts], Acc) ->
+    normparts(RestParts, [Part | Acc]).
+
 % works like list_to_existing_atom, except can be list or binary and it
 % gives you the original value instead of an error if no existing atom.
-to_existing_atom(V) when is_list(V)->
+to_existing_atom(V) when is_list(V) ->
     try list_to_existing_atom(V) catch _ -> V end;
-to_existing_atom(V) when is_binary(V)->
+to_existing_atom(V) when is_binary(V) ->
     try list_to_existing_atom(?b2l(V)) catch _ -> V end;
-to_existing_atom(V) when is_atom(V)->
+to_existing_atom(V) when is_atom(V) ->
     V.
 
 
@@ -55,9 +81,6 @@ terminate_linked(Reason) ->
     ok.
 
 
-new_uuid() ->
-    list_to_binary(to_hex(crypto:rand_bytes(16))).
-
 to_hex([]) ->
     [];
 to_hex(Bin) when is_binary(Bin) ->
@@ -69,7 +92,7 @@ to_digit(N) when N < 10 -> $0 + N;
 to_digit(N)             -> $a + N-10.
 
 
-parse_term(Bin) when is_binary(Bin)->
+parse_term(Bin) when is_binary(Bin) ->
     parse_term(binary_to_list(Bin));
 parse_term(List) ->
     {ok, Tokens, _} = erl_scan:string(List ++ "."),
@@ -141,14 +164,14 @@ ascii_lower([], Acc) ->
     lists:reverse(Acc);
 ascii_lower([Char | RestString], Acc) when Char >= $A, Char =< $B ->
     ascii_lower(RestString, [Char + ($a-$A) | Acc]);
-ascii_lower([Char | RestString], Acc)->
+ascii_lower([Char | RestString], Acc) ->
     ascii_lower(RestString, [Char | Acc]).
 
 % Is a character whitespace?
-is_whitespace($\s)-> true;
-is_whitespace($\t)-> true;
-is_whitespace($\n)-> true;
-is_whitespace($\r)-> true;
+is_whitespace($\s) -> true;
+is_whitespace($\t) -> true;
+is_whitespace($\n) -> true;
+is_whitespace($\r) -> true;
 is_whitespace(_Else) -> false.
 
 
@@ -183,7 +206,7 @@ implode([H|T], Sep, Acc) ->
 drv_port() ->
     case get(couch_drv_port) of
     undefined ->
-        Port = open_port({spawn, "couch_erl_driver"}, []),
+        Port = open_port({spawn, "couch_icu_driver"}, []),
         put(couch_drv_port, Port),
         Port;
     Port ->
@@ -199,9 +222,8 @@ collate(A, B, Options) when is_binary(A), is_binary(B) ->
         true -> 1; % Case insensitive
         false -> 0 % Case sensitive
     end,
-
-    SizeA = size(A),
-    SizeB = size(B),
+    SizeA = byte_size(A),
+    SizeB = byte_size(B),
     Bin = <<SizeA:32/native, A/binary, SizeB:32/native, B/binary>>,
     [Result] = erlang:port_control(drv_port(), Operation, Bin),
     % Result is 0 for lt, 1 for eq and 2 for gt. Subtract 1 to return the
@@ -220,9 +242,7 @@ should_flush(MemThreshHold) ->
         {memory, ProcMem2} = process_info(self(), memory),
         BinMem2 = lists:foldl(fun({_Id, Size, _NRefs}, Acc) -> Size+Acc end,
             0, element(2,process_info(self(), binary))),
-        if ProcMem2+BinMem2 > MemThreshHold ->
-            true;
-        true -> false end;
+        ProcMem2+BinMem2 > MemThreshHold;
     true -> false end.
 
 
@@ -280,7 +300,7 @@ encodeBase64Url(<<>>, Acc) ->
 %%
 %% decodeBase64(BinaryChars) -> Binary
 %%
-decodeBase64(Cs) when is_list(Cs)->
+decodeBase64(Cs) when is_list(Cs) ->
     decodeBase64(list_to_binary(Cs));
 decodeBase64(Cs) ->
     decode1(Cs, <<>>).
@@ -297,7 +317,7 @@ decode1(<<C1, C2, C3, C4, Cs/binary>>, Acc) ->
 decode1(<<>>, Acc) ->
     Acc.
 
-decodeBase64Url(Cs) when is_list(Cs)->
+decodeBase64Url(Cs) when is_list(Cs) ->
     decodeBase64Url(list_to_binary(Cs));
 decodeBase64Url(Cs) ->
     decode1Url(Cs, <<>>).
@@ -364,6 +384,13 @@ to_binary(V) when is_atom(V) ->
 to_binary(V) ->
     list_to_binary(io_lib:format("~p", [V])).
 
+to_integer(V) when is_integer(V) ->
+    V;
+to_integer(V) when is_list(V) ->
+    erlang:list_to_integer(V);
+to_integer(V) when is_binary(V) ->
+    erlang:list_to_integer(binary_to_list(V)).
+
 to_list(V) when is_list(V) ->
     V;
 to_list(V) when is_binary(V) ->
@@ -394,4 +421,58 @@ url_encode([H|T]) ->
         end
     end;
 url_encode([]) ->
-    [].
\ No newline at end of file
+    [].
+
+json_encode(V) ->
+    Handler =
+    fun({L}) when is_list(L) ->
+        {struct,L};
+    (Bad) ->
+        exit({json_encode, {bad_term, Bad}})
+    end,
+    (mochijson2:encoder([{handler, Handler}]))(V).
+
+json_decode(V) ->
+    try (mochijson2:decoder([{object_hook, fun({struct,L}) -> {L} end}]))(V)
+    catch
+        _Type:_Error ->
+            throw({invalid_json,V})
+    end.
+
+verify([X|RestX], [Y|RestY], Result) ->
+    verify(RestX, RestY, (X bxor Y) bor Result);
+verify([], [], Result) ->
+    Result == 0.
+
+verify(<<X/binary>>, <<Y/binary>>) ->
+    verify(?b2l(X), ?b2l(Y));
+verify(X, Y) when is_list(X) and is_list(Y) ->
+    case length(X) == length(Y) of
+        true ->
+            verify(X, Y, 0);
+        false ->
+            false
+    end;
+verify(_X, _Y) -> false.
+
+compressible_att_type(MimeType) when is_binary(MimeType) ->
+    compressible_att_type(?b2l(MimeType));
+compressible_att_type(MimeType) ->
+    TypeExpList = re:split(
+        couch_config:get("attachments", "compressible_types", ""),
+        "\\s+",
+        [{return, list}]
+    ),
+    lists:any(
+        fun(TypeExp) ->
+            Regexp = "^\\s*" ++
+                re:replace(TypeExp, "\\*", ".*", [{return, list}]) ++ "\\s*$",
+            case re:run(MimeType, Regexp, [caseless]) of
+            {match, _} ->
+                true;
+            _ ->
+                false
+            end
+        end,
+        [T || T <- TypeExpList, T /= []]
+    ).
