@@ -13,10 +13,10 @@
 -module(couch_view).
 -behaviour(gen_server).
 
--export([start_link/0,fold/4,fold/5,less_json/2,less_json_keys/2,expand_dups/2,
+-export([start_link/0,fold/4,less_json/2,less_json_ids/2,expand_dups/2,
     detuple_kvs/2,init/1,terminate/2,handle_call/3,handle_cast/2,handle_info/2,
     code_change/3,get_reduce_view/4,get_temp_reduce_view/5,get_temp_map_view/4,
-    get_map_view/4,get_row_count/1,reduce_to_count/1,fold_reduce/7,
+    get_map_view/4,get_row_count/1,reduce_to_count/1,fold_reduce/4,
     extract_map_view/1,get_group_server/2,get_group_info/2,cleanup_index_files/1]).
 
 -include("couch_db.hrl").
@@ -89,9 +89,9 @@ cleanup_index_files(Db) ->
     RegExp = "("++ string:join(Sigs, "|") ++")",
 
     % filter out the ones in use
-    DeleteFiles = lists:filter(fun(FilePath) ->
-            re:run(FilePath, RegExp, [{capture, none}]) == nomatch
-        end, FileList),
+    DeleteFiles = [FilePath
+		   || FilePath <- FileList,
+		      re:run(FilePath, RegExp, [{capture, none}]) =:= nomatch],
     % delete unused files
     ?LOG_DEBUG("deleting unused view index files: ~p",[DeleteFiles]),
     [file:delete(File)||File <- DeleteFiles],
@@ -152,16 +152,14 @@ expand_dups([{Key, {dups, Vals}} | Rest], Acc) ->
 expand_dups([KV | Rest], Acc) ->
     expand_dups(Rest, [KV | Acc]).
 
-fold_reduce({temp_reduce, #view{btree=Bt}}, Dir, StartKey, EndKey, GroupFun, Fun, Acc) ->
-
+fold_reduce({temp_reduce, #view{btree=Bt}}, Fun, Acc, Options) ->
     WrapperFun = fun({GroupedKey, _}, PartialReds, Acc0) ->
             {_, [Red]} = couch_btree:final_reduce(Bt, PartialReds),
             Fun(GroupedKey, Red, Acc0)
         end,
-    couch_btree:fold_reduce(Bt, Dir, StartKey, EndKey, GroupFun,
-            WrapperFun, Acc);
+    couch_btree:fold_reduce(Bt, WrapperFun, Acc, Options);
 
-fold_reduce({reduce, NthRed, Lang, #view{btree=Bt, reduce_funs=RedFuns}}, Dir, StartKey, EndKey, GroupFun, Fun, Acc) ->
+fold_reduce({reduce, NthRed, Lang, #view{btree=Bt, reduce_funs=RedFuns}}, Fun, Acc, Options) ->
     PreResultPadding = lists:duplicate(NthRed - 1, []),
     PostResultPadding = lists:duplicate(length(RedFuns) - NthRed, []),
     {_Name, FunSrc} = lists:nth(NthRed,RedFuns),
@@ -178,8 +176,7 @@ fold_reduce({reduce, NthRed, Lang, #view{btree=Bt, reduce_funs=RedFuns}}, Dir, S
             {_, Reds} = couch_btree:final_reduce(ReduceFun, PartialReds),
             Fun(GroupedKey, lists:nth(NthRed, Reds), Acc0)
         end,
-    couch_btree:fold_reduce(Bt, Dir, StartKey, EndKey, GroupFun,
-            WrapperFun, Acc).
+    couch_btree:fold_reduce(Bt, WrapperFun, Acc, Options).
 
 get_key_pos(_Key, [], _N) ->
     0;
@@ -239,15 +236,13 @@ fold_fun(Fun, [KV|Rest], {KVReds, Reds}, Acc) ->
         {stop, Acc2}
     end.
 
-fold(#view{btree=Btree}, Dir, Fun, Acc) ->
-    fold(Btree, nil, Dir, Fun, Acc).
 
-fold(#view{btree=Btree}, StartKey, Dir, Fun, Acc) ->
+fold(#view{btree=Btree}, Fun, Acc, Options) ->
     WrapperFun =
         fun(KV, Reds, Acc2) ->
             fold_fun(Fun, expand_dups([KV],[]), Reds, Acc2)
         end,
-    {ok, _AccResult} = couch_btree:fold(Btree, StartKey, Dir, WrapperFun, Acc).
+    {ok, _LastReduce, _AccResult} = couch_btree:fold(Btree, WrapperFun, Acc, Options).
 
 
 init([]) ->
@@ -289,6 +284,15 @@ handle_call({get_group_server, DbName,
         {ok, NewPid} ->
             add_to_ets(NewPid, DbName, Sig),
             {reply, {ok, NewPid}, Server};
+        {error, invalid_view_seq} ->
+            do_reset_indexes(DbName, Root),
+            case (catch couch_view_group:start_link({Root, DbName, Group})) of
+            {ok, NewPid} ->
+                add_to_ets(NewPid, DbName, Sig),
+                {reply, {ok, NewPid}, Server};
+            Error ->
+                {reply, Error, Server}
+            end;
         Error ->
             {reply, Error, Server}
         end;
@@ -297,6 +301,10 @@ handle_call({get_group_server, DbName,
     end.
 
 handle_cast({reset_indexes, DbName}, #server{root_dir=Root}=Server) ->
+    do_reset_indexes(DbName, Root),
+    {noreply, Server}.
+
+do_reset_indexes(DbName, Root) ->
     % shutdown all the updaters and clear the files, the db got changed
     Names = ets:lookup(couch_groups_by_db, DbName),
     lists:foreach(
@@ -309,8 +317,7 @@ handle_cast({reset_indexes, DbName}, #server{root_dir=Root}=Server) ->
             end
         end, Names),
     delete_index_dir(Root, DbName),
-    file:delete(Root ++ "/." ++ binary_to_list(DbName) ++ "_temp"),
-    {noreply, Server}.
+    file:delete(Root ++ "/." ++ ?b2l(DbName) ++ "_temp").
 
 handle_info({'EXIT', FromPid, Reason}, Server) ->
     case ets:lookup(couch_groups_by_updater, FromPid) of
@@ -360,19 +367,24 @@ nuke_dir(Dir) ->
         ok = file:del_dir(Dir)
     end.
 
+
 % keys come back in the language of btree - tuples.
-less_json_keys(A, B) ->
-    less_json(tuple_to_list(A), tuple_to_list(B)).
+less_json_ids({JsonA, IdA}, {JsonB, IdB}) ->
+    case JsonA == JsonB of
+    false ->
+        less_json(JsonA, JsonB);
+    true ->
+        IdA < IdB
+    end.
+        
 
 less_json(A, B) ->
     TypeA = type_sort(A),
     TypeB = type_sort(B),
-    if
-    TypeA == TypeB ->
-        Less = less_same_type(A,B),
-        Less;
-    true ->
-        TypeA < TypeB
+    if TypeA == TypeB ->
+	    less_same_type(A, B);
+       true ->
+	    TypeA < TypeB
     end.
 
 type_sort(V) when is_atom(V) -> 0;
@@ -384,7 +396,6 @@ type_sort({V}) when is_list(V) -> 4;
 type_sort(V) when is_tuple(V) -> 5.
 
 
-atom_sort(nil) -> 0;
 atom_sort(null) -> 1;
 atom_sort(false) -> 2;
 atom_sort(true) -> 3.
