@@ -78,17 +78,25 @@ start_link(InitArgs) ->
 init({InitArgs, ReturnPid, Ref}) ->
     process_flag(trap_exit, true),
     case prepare_group(InitArgs, false) of
-    {ok, #group{db=Db, fd=Fd}=Group} ->
-        couch_db:monitor(Db),
-        Owner = self(),
-        Pid = spawn_link(fun()-> couch_view_updater:update(Owner, Group) end),
-        {ok, RefCounter} = couch_ref_counter:start([Fd]),
-        {ok, #group_state{
-                db_name=couch_db:name(Db),
-                init_args=InitArgs,
-                updater_pid = Pid,
-                group=Group,
-                ref_counter=RefCounter}};
+    {ok, #group{db=Db, fd=Fd, current_seq=Seq}=Group} ->
+        case Seq > couch_db:get_update_seq(Db) of
+        true ->
+            ReturnPid ! {Ref, self(), {error, invalid_view_seq}},
+            ignore;
+        _ ->
+            couch_db:monitor(Db),
+            Owner = self(),
+            Pid = spawn_link(
+                fun()-> couch_view_updater:update(Owner, Group) end
+            ),
+            {ok, RefCounter} = couch_ref_counter:start([Fd]),
+            {ok, #group_state{
+                    db_name=couch_db:name(Db),
+                    init_args=InitArgs,
+                    updater_pid = Pid,
+                    group=Group,
+                    ref_counter=RefCounter}}
+        end;
     Error ->
         ReturnPid ! {Ref, self(), Error},
         ignore
@@ -117,7 +125,7 @@ handle_call({request_group, RequestSeq}, From,
             updater_pid=nil,
             waiting_list=WaitList
             }=State) when RequestSeq > Seq ->
-    {ok, Db} = couch_db:open(DbName, []),
+    {ok, Db} = couch_db:open_int(DbName, []),
     Group2 = Group#group{db=Db},
     Owner = self(),
     Pid = spawn_link(fun()-> couch_view_updater:update(Owner, Group2) end),
@@ -145,11 +153,8 @@ handle_call({request_group, RequestSeq}, From,
         waiting_list=[{From, RequestSeq}|WaitList]
         }, infinity};
 
-handle_call(request_group_info, _From, #group_state{
-            group = Group,
-            compactor_pid = CompactorPid
-        } = State) ->
-    GroupInfo = get_group_info(Group, CompactorPid),
+handle_call(request_group_info, _From, State) ->
+    GroupInfo = get_group_info(State),
     {reply, {ok, GroupInfo}, State}.
 
 handle_cast({start_compact, CompactFun}, #group_state{compactor_pid=nil}
@@ -159,7 +164,7 @@ handle_cast({start_compact, CompactFun}, #group_state{compactor_pid=nil}
         init_args = {RootDir, DbName, _}
     } = State,
     ?LOG_INFO("View index compaction starting for ~s ~s", [DbName, GroupId]),
-    {ok, Db} = couch_db:open(DbName, []),
+    {ok, Db} = couch_db:open_int(DbName, []),
     {ok, Fd} = open_index_file(compact, RootDir, DbName, GroupSig),
     NewGroup = reset_file(Db, Fd, DbName, Group),
     Pid = spawn_link(fun() -> CompactFun(Group, NewGroup) end),
@@ -219,7 +224,7 @@ handle_cast({compact_done, NewGroup}, State) ->
     ?LOG_INFO("View index compaction still behind for ~s ~s -- current: ~p " ++
         "compact: ~p", [DbName, GroupId, CurrentSeq, NewGroup#group.current_seq]),
     couch_db:close(NewGroup#group.db),
-    {ok, Db} = couch_db:open(DbName, []),
+    {ok, Db} = couch_db:open_int(DbName, []),
     Pid = spawn_link(fun() ->
         {_,Ref} = erlang:spawn_monitor(fun() ->
             couch_view_updater:update(nil, NewGroup#group{db = Db})
@@ -252,7 +257,7 @@ handle_cast({partial_update, _, _}, State) ->
     {noreply, State}.
 
 handle_info(delayed_commit, #group_state{db_name=DbName,group=Group}=State) ->
-    {ok, Db} = couch_db:open(DbName, []),
+    {ok, Db} = couch_db:open_int(DbName, []),
     CommittedSeq = couch_db:get_committed_update_seq(Db),
     couch_db:close(Db),
     if CommittedSeq >= Group#group.current_seq ->
@@ -287,7 +292,7 @@ handle_info({'EXIT', FromPid, {new_group, #group{db=Db}=Group}},
                 group=Group#group{db=nil}, updater_pid=nil}};
     StillWaiting ->
         % we still have some waiters, reopen the database and reupdate the index
-        {ok, Db2} = couch_db:open(DbName, []),
+        {ok, Db2} = couch_db:open_int(DbName, []),
         Group2 = Group#group{db=Db2},
         Owner = self(),
         Pid = spawn_link(fun() -> couch_view_updater:update(Owner, Group2) end),
@@ -367,7 +372,7 @@ reply_all(#group_state{waiting_list=WaitList}=State, Reply) ->
     State#group_state{waiting_list=[]}.
 
 prepare_group({RootDir, DbName, #group{sig=Sig}=Group}, ForceReset)->
-    case couch_db:open(DbName, []) of
+    case couch_db:open_int(DbName, []) of
     {ok, Db} ->
         case open_index_file(RootDir, DbName, Sig) of
         {ok, Fd} ->
@@ -432,28 +437,29 @@ open_index_file(compact, RootDir, DbName, GroupSig) ->
     end.
 
 open_temp_group(DbName, Language, DesignOptions, MapSrc, RedSrc) ->
-    case couch_db:open(DbName, []) of
+    case couch_db:open_int(DbName, []) of
     {ok, Db} ->
         View = #view{map_names=[<<"_temp">>],
             id_num=0,
             btree=nil,
             def=MapSrc,
-            reduce_funs= if RedSrc==[] -> []; true -> [{<<"_temp">>, RedSrc}] end},
+            reduce_funs= if RedSrc==[] -> []; true -> [{<<"_temp">>, RedSrc}] end,
+            options=DesignOptions},
 
-        {ok, Db, #group{
-            name = <<"_temp">>,
-            db=Db,
-            views=[View],
-            def_lang=Language,
-            design_options=DesignOptions,
-            sig = erlang:md5(term_to_binary({[View], Language, DesignOptions}))
-        }};
+        {ok, Db, set_view_sig(#group{name = <<"_temp">>, db=Db, views=[View],
+            def_lang=Language, design_options=DesignOptions})};
     Error ->
         Error
     end.
 
+set_view_sig(#group{
+            views=Views,
+            def_lang=Language,
+            design_options=DesignOptions}=G) ->
+    G#group{sig=erlang:md5(term_to_binary({Views, Language, DesignOptions}))}.
+
 open_db_group(DbName, GroupId) ->
-    case couch_db:open(DbName, []) of
+    case couch_db:open_int(DbName, []) of
     {ok, Db} ->
         case couch_db:open_doc(Db, GroupId) of
         {ok, Doc} ->
@@ -466,17 +472,32 @@ open_db_group(DbName, GroupId) ->
         Else
     end.
 
-get_group_info(#group{
+get_group_info(State) ->
+    #group_state{
+        group=Group,
+        updater_pid=UpdaterPid,
+        compactor_pid=CompactorPid,
+        waiting_commit=WaitingCommit,
+        waiting_list=WaitersList
+    } = State,
+    #group{
         fd = Fd,
         sig = GroupSig,
-        def_lang = Lang
-    }, CompactorPid) ->
+        def_lang = Lang,
+        current_seq=CurrentSeq,
+        purge_seq=PurgeSeq
+    } = Group,
     {ok, Size} = couch_file:bytes(Fd),
     [
         {signature, ?l2b(hex_sig(GroupSig))},
         {language, Lang},
         {disk_size, Size},
-        {compact_running, CompactorPid /= nil}
+        {updater_running, UpdaterPid /= nil},
+        {compact_running, CompactorPid /= nil},
+        {waiting_commit, WaitingCommit},
+        {waiting_clients, length(WaitersList)},
+        {update_seq, CurrentSeq},
+        {purge_seq, PurgeSeq}
     ].
 
 % maybe move to another module
@@ -484,20 +505,17 @@ design_doc_to_view_group(#doc{id=Id,body={Fields}}) ->
     Language = proplists:get_value(<<"language">>, Fields, <<"javascript">>),
     {DesignOptions} = proplists:get_value(<<"options">>, Fields, {[]}),
     {RawViews} = proplists:get_value(<<"views">>, Fields, {[]}),
-    % sort the views by name to avoid spurious signature changes
-    SortedRawViews = lists:sort(fun({Name1, _}, {Name2, _}) ->
-            Name1 >= Name2
-        end, RawViews),
     % add the views to a dictionary object, with the map source as the key
     DictBySrc =
     lists:foldl(
         fun({Name, {MRFuns}}, DictBySrcAcc) ->
             MapSrc = proplists:get_value(<<"map">>, MRFuns),
             RedSrc = proplists:get_value(<<"reduce">>, MRFuns, null),
+            {ViewOptions} = proplists:get_value(<<"options">>, MRFuns, {[]}),
             View =
-            case dict:find(MapSrc, DictBySrcAcc) of
+            case dict:find({MapSrc, ViewOptions}, DictBySrcAcc) of
                 {ok, View0} -> View0;
-                error -> #view{def=MapSrc} % create new view object
+                error -> #view{def=MapSrc, options=ViewOptions} % create new view object
             end,
             View2 =
             if RedSrc == null ->
@@ -505,16 +523,15 @@ design_doc_to_view_group(#doc{id=Id,body={Fields}}) ->
             true ->
                 View#view{reduce_funs=[{Name,RedSrc}|View#view.reduce_funs]}
             end,
-            dict:store(MapSrc, View2, DictBySrcAcc)
-        end, dict:new(), SortedRawViews),
+            dict:store({MapSrc, ViewOptions}, View2, DictBySrcAcc)
+        end, dict:new(), RawViews),
     % number the views
     {Views, _N} = lists:mapfoldl(
         fun({_Src, View}, N) ->
             {View#view{id_num=N},N+1}
-        end, 0, dict:to_list(DictBySrc)),
+        end, 0, lists:sort(dict:to_list(DictBySrc))),
 
-    Group = #group{name=Id, views=Views, def_lang=Language, design_options=DesignOptions},
-    Group#group{sig=erlang:md5(term_to_binary({Views, Language, DesignOptions}))}.
+    set_view_sig(#group{name=Id, views=Views, def_lang=Language, design_options=DesignOptions}).
 
 reset_group(#group{views=Views}=Group) ->
     Views2 = [View#view{btree=nil} || View <- Views],
@@ -534,12 +551,13 @@ init_group(Db, Fd, #group{views=Views}=Group, nil) ->
     init_group(Db, Fd, Group,
         #index_header{seq=0, purge_seq=couch_db:get_purge_seq(Db),
             id_btree_state=nil, view_states=[nil || _ <- Views]});
-init_group(Db, Fd, #group{def_lang=Lang,views=Views}=Group, IndexHeader) ->
+init_group(Db, Fd, #group{def_lang=Lang,views=Views}=
+            Group, IndexHeader) ->
      #index_header{seq=Seq, purge_seq=PurgeSeq,
             id_btree_state=IdBtreeState, view_states=ViewStates} = IndexHeader,
     {ok, IdBtree} = couch_btree:open(IdBtreeState, Fd),
     Views2 = lists:zipwith(
-        fun(BtreeState, #view{reduce_funs=RedFuns}=View) ->
+        fun(BtreeState, #view{reduce_funs=RedFuns,options=Options}=View) ->
             FunSrcs = [FunSrc || {_Name, FunSrc} <- RedFuns],
             ReduceFun =
                 fun(reduce, KVs) ->
@@ -555,8 +573,15 @@ init_group(Db, Fd, #group{def_lang=Lang,views=Views}=Group, IndexHeader) ->
                         UserReds),
                     {Count, Reduced}
                 end,
+            
+            case proplists:get_value(<<"collation">>, Options, <<"default">>) of
+            <<"default">> ->
+                Less = fun couch_view:less_json_ids/2;
+            <<"raw">> ->
+                Less = fun(A,B) -> A < B end
+            end,
             {ok, Btree} = couch_btree:open(BtreeState, Fd,
-                        [{less, fun couch_view:less_json_keys/2},
+                        [{less, Less},
                             {reduce, ReduceFun}]),
             View#view{btree=Btree}
         end,
